import json
from typing import Dict, Iterable, Optional

from cytoolz import reduceby
from requests import Response

from blockapi.v2.base import (
    ApiException,
    ApiOptions,
    CustomizableBlockchainApi,
    IBalance,
    InvalidAddressException,
)
from blockapi.v2.coins import COIN_SOL
from blockapi.v2.models import BalanceItem, Blockchain, Coin, CoinInfo


class SolanaApi(CustomizableBlockchainApi, IBalance):
    """
    Solana RPC
    API docs: https://docs.solana.com/apps/jsonrpc-api
    """

    API_BASE_URL = 'https://api.mainnet-beta.solana.com/'

    coin = COIN_SOL
    api_options = ApiOptions(
        blockchain=Blockchain.SOLANA,
        base_url=API_BASE_URL,
        start_offset=0,
        max_items_per_page=1000,
        page_offset_step=1,
    )

    # API uses post requests
    supported_requests = {}

    token_program_id = 'TokenkegQfeZyiNwAJbNbGKPFXCWuBvf9Ss623VQ5DA'
    _tokens_map: Optional[Dict[str, Dict]] = None

    @property
    def tokens_map(self) -> Dict[str, Dict]:
        if self._tokens_map is None:
            response = self._session.get(
                'https://raw.githubusercontent.com/solana-labs/token-list/'
                'main/src/tokens/solana.tokenlist.json'
            )
            token_list = response.json()
            self._tokens_map = {t['address']: t for t in token_list['tokens']}

        return self._tokens_map

    def get_balance(self, address: str):
        balances = []

        sol_balance = self._get_sol_balance(address)
        if sol_balance is not None:
            balances.append(sol_balance)

        token_balances = list(self._yield_token_balances(address))
        if token_balances:
            merged_token_balances = self.merge_balances_with_same_coin(token_balances)
            balances.extend(merged_token_balances)

        return balances

    @staticmethod
    def merge_balances_with_same_coin(
        token_balances: list[BalanceItem],
    ) -> list[BalanceItem]:
        return list(
            reduceby(
                key=lambda b: b.coin.address,
                binop=lambda v1, v2: v1 + v2,
                seq=token_balances,
            ).values()
        )

    def _get_sol_balance(
        self,
        address: str,
    ) -> Optional[BalanceItem]:
        response = self._request(method='getBalance', params=[address])
        if int(response['result']['value']) == 0:
            return

        return BalanceItem.from_api(
            balance_raw=response['result']['value'],
            coin=self.coin,
            last_updated=None,
            raw=response,
        )

    def _yield_token_balances(self, address: str) -> Iterable[BalanceItem]:
        response = self._request(
            method='getTokenAccountsByOwner',
            params=[
                address,
                {'programId': self.token_program_id},
                {'encoding': 'jsonParsed'},
            ],
        )

        for raw_balance in response['result']['value']:
            balance = self._parse_token_balance(raw_balance)
            if balance is not None:
                yield balance

    def _parse_token_balance(self, raw: Dict) -> Optional[BalanceItem]:
        info = raw['account']['data']['parsed']['info']
        if int(info['tokenAmount']['amount']) == 0:
            return

        address = info['mint']

        # TODO unknown token is for 99% NFT, add loading NFT metadata using metaplex
        #  token account
        if address not in self.tokens_map:
            return

        return BalanceItem.from_api(
            balance_raw=info['tokenAmount']['amount'],
            coin=self._get_token_data(address),
            raw=raw,
        )

    def _get_token_data(self, address: str) -> Coin:
        raw_token = self.tokens_map[address]
        extensions = raw_token.get('extensions', {})

        return Coin(
            symbol=raw_token['symbol'],
            name=raw_token['name'],
            decimals=raw_token['decimals'],
            blockchain=Blockchain.SOLANA,
            address=address,
            standards=['SPL'],
            info=CoinInfo.from_api(
                tags=raw_token.get('tags'),
                logo_url=raw_token.get('logoURI'),
                # TODO sometimes coingeckoId is not mapped correctly
                # coingecko_id=extensions.get('coingeckoId'),
                website=extensions.get('website'),
            ),
        )

    def _request(self, method, params):
        body = json.dumps(
            {'jsonrpc': '2.0', 'id': 1, 'method': method, 'params': params}
        )
        return self.post(body=body, headers={'Content-Type': 'application/json'})

    def _opt_raise_on_other_error(self, response: Response) -> None:
        json_response = response.json()
        if 'error' not in json_response:
            return

        if 'Invalid param' in json_response['error']['message']:
            raise InvalidAddressException(f'Invalid address format.')
        else:
            raise ApiException(json_response['error']['message'])



import re

from blockapi.api import SolanaApi
from blockapi.test.v2.test_data import (
    get_debank_addresses,
    yield_api_ibalance_classes,
    yield_covalent_api_classes,
)
from blockapi.v2.api.covalenth.arbitrum import ArbitrumCovalentApi
from blockapi.v2.api.covalenth.astar import AstarCovalentApi
from blockapi.v2.api.covalenth.avalanche import AvalancheCovalentApi
from blockapi.v2.api.covalenth.axie import AxieCovalentApi
from blockapi.v2.api.covalenth.binance_smart_chain import BscCovalentApi
from blockapi.v2.api.covalenth.ethereum import EthCovalentApi
from blockapi.v2.api.covalenth.fantom import FantomCovalentApi
from blockapi.v2.api.covalenth.heco import HECOCovalentApi
from blockapi.v2.api.covalenth.iotex import IoTEXCovalentApi
from blockapi.v2.api.covalenth.klaytn import KlaytnCovalentApi
from blockapi.v2.api.covalenth.moonbeam import MoonBeamCovalentApi
from blockapi.v2.api.covalenth.palm import PalmCovalentApi
from blockapi.v2.api.covalenth.polygon import PolygonCovalentApi
from blockapi.v2.api.covalenth.rsk import RskCovalentApi
from blockapi.v2.api.debank import DebankApi
from blockapi.v2.api.ethplorer import EthplorerApi
from blockapi.v2.api.perpetual import PerpetualApi
from blockapi.v2.api.synthetix import SynthetixApi


def test_enumerate_subclasses():
    classes = [x.__name__ for x in yield_api_ibalance_classes()]

    assert classes
    assert SolanaApi.__name__ in classes
    assert EthplorerApi.__name__ in classes
    assert DebankApi.__name__ not in classes
    assert PerpetualApi.__name__ not in classes
    assert SynthetixApi.__name__ not in classes

    for cls in classes:
        assert 'Base' not in cls


def test_enumerate_covalent_all_classes():
    assert [
        ArbitrumCovalentApi,
        AstarCovalentApi,
        AvalancheCovalentApi,
        AxieCovalentApi,
        BscCovalentApi,
        FantomCovalentApi,
        HECOCovalentApi,
        IoTEXCovalentApi,
        KlaytnCovalentApi,
        MoonBeamCovalentApi,
        PalmCovalentApi,
        PolygonCovalentApi,
        RskCovalentApi,
        EthCovalentApi,
    ] == yield_covalent_api_classes()


def test_yield_debank_address():
    for param in get_debank_addresses():
        for address in param.values:
            assert re.match('^0x[a-fA-F0-9]{40}$', address)



from blockapi.v2.api.covalenth.base import CovalentApiBase
from blockapi.v2.base import ApiOptions
from blockapi.v2.coins import COIN_BNB
from blockapi.v2.models import Blockchain


class BscCovalentApi(CovalentApiBase):
    CHAIN_ID = 56

    api_options = ApiOptions(
        blockchain=Blockchain.BINANCE_SMART_CHAIN,
        base_url=CovalentApiBase.API_BASE_URL,
        rate_limit=CovalentApiBase.API_BASE_RATE_LIMIT,
    )

    coin = COIN_BNB



from typing import Iterable, List, Optional

from blockapi.v2.base import BlockchainApi, IBalance
from blockapi.v2.coins import COIN_BOS
from blockapi.v2.models import (
    ApiOptions,
    AssetType,
    BalanceItem,
    Blockchain,
    OperationDirection,
    OperationItem,
    OperationType,
    TransactionItem,
    TransactionStatus,
)


class BlockchainosApi(BlockchainApi, IBalance):
    coin = COIN_BOS
    api_options = ApiOptions(
        blockchain=Blockchain.BOS,
        base_url='https://mainnet.blockchainos.org',
        rate_limit=0,
    )

    supported_requests = {
        'get_balance': '/api/v1/accounts/{address}',
        'get_txs': '/api/v1/accounts/{address}/transactions?limit={limit}&reverse=true',
        'get_ops': '/api/v1/transactions/{hash}/operations/{index}',
    }

    def get_balance(self, address: str) -> List[BalanceItem]:
        raw_balances = self.get('get_balance', address=address)
        return list(self._parse_balances(raw_balances))

    def get_transactions(
        self, address: str, *, limit: int = 10
    ) -> List[TransactionItem]:
        raw_transactions = self.get('get_txs', address=address, limit=limit)
        return list(self._parse_transactions(address, raw_transactions))

    def get_operations(
        self, address: str, hash_: str, count: int
    ) -> List[OperationItem]:
        ops = []
        for index in range(count):
            raw_operations = self.get('get_ops', hash=hash_, index=index)
            ops.extend(self._parse_operations(address, raw_operations))

        return ops

    def _parse_balances(self, raw_balances: dict) -> Iterable[BalanceItem]:
        yield BalanceItem.from_api(
            balance_raw=raw_balances.get('balance'),
            coin=self.coin,
            asset_type=AssetType.AVAILABLE,
            raw=raw_balances,
        )

    def _parse_transactions(
        self, address: str, raw_transactions: dict
    ) -> Iterable[TransactionItem]:
        records = self._get_records(raw_transactions)
        if not records:
            return

        for tx in records:
            hash_ = tx.get('hash')
            op_count = int(tx.get('operation_count'))
            yield TransactionItem.from_api(
                fee_raw=tx.get('fee'),
                coin=self.coin,
                date=tx.get('created'),
                hash=hash_,
                status=TransactionStatus.CONFIRMED,
                operations=self.get_operations(address, hash_, op_count),
                raw=raw_transactions,
            )

    def _parse_operations(self, address: str, raw_op: dict):
        body = raw_op.get('body')
        from_address = raw_op.get('source')
        yield OperationItem.from_api(
            amount_raw=body.get('amount'),
            coin=self.coin,
            from_address=from_address,
            to_address=raw_op.get('target'),
            hash=raw_op.get('tx_hash'),
            type=OperationType(raw_op.get('type')),
            direction=OperationDirection.OUTGOING
            if address == from_address
            else OperationDirection.INCOMING,
            raw=raw_op,
            confirmed=raw_op.get('confirmed'),
        )

    @staticmethod
    def _get_records(data: dict) -> Optional[dict]:
        embedded = data.get('_embedded')
        if not embedded:
            return None

        return embedded.get('records')



from decimal import Decimal

import pytest

from blockapi.v2.api.solana import SolanaApi
from blockapi.v2.models import AssetType, BalanceItem, Blockchain, Coin, CoinInfo


@pytest.fixture
def solana_api():
    return SolanaApi()


@pytest.fixture()
def balances_with_same_coin():
    return [
        BalanceItem(
            balance=Decimal('1'),
            balance_raw=Decimal('1'),
            raw={'raw1': 'raw_value_1'},
            coin=Coin(
                symbol='unknown',
                name='unknown',
                decimals=0,
                blockchain=Blockchain.SOLANA,
                address='HEL6KGUEvwYgTtcjenf9qeAb2Zg9Yr77usWPY9UZvoQj',
                standards=None,
                protocol_id=None,
                info=None,
            ),
            asset_type=AssetType.AVAILABLE,
            last_updated=None,
            protocol=None,
            is_wallet=True,
        ),
        BalanceItem(
            balance=Decimal('11'),
            balance_raw=Decimal('11'),
            raw={'raw2': 'raw_value_2'},
            coin=Coin(
                symbol='unknown',
                name='unknown',
                decimals=0,
                blockchain=Blockchain.SOLANA,
                address='HEL6KGUEvwYgTtcjenf9qeAb2Zg9Yr77usWPY9UZvoQj',
                standards=None,
                protocol_id=None,
                info=None,
            ),
            asset_type=AssetType.AVAILABLE,
            last_updated=None,
            protocol=None,
            is_wallet=True,
        ),
    ]


def test_merge_balances_with_same_coin(solana_api, balances_with_same_coin):
    merged = solana_api.merge_balances_with_same_coin(balances_with_same_coin)
    assert len(merged) == 1

    merged = merged[0]
    assert merged.balance == 12
    assert merged.balance_raw == 12
    assert merged.raw == {
        "merged": [balances_with_same_coin[0].raw, balances_with_same_coin[1].raw]
    }


@pytest.fixture()
def balances_with_different_coins():
    return [
        BalanceItem(
            balance=Decimal('1'),
            balance_raw=Decimal('1'),
            raw={
                'account': {
                    'data': {
                        'parsed': {
                            'info': {
                                'isNative': False,
                                'mint': 'HEL6KGUEvwYgTtcjenf9qeAb2Zg9Yr77usWPY9UZvoQj',
                                'owner': 'FEeSRuEDk8ENZbpzXjn4uHPz3LQijbeKRzhqVr5zPSJ9',
                                'state': 'initialized',
                                'tokenAmount': {
                                    'amount': '1',
                                    'decimals': 0,
                                    'uiAmount': 1.0,
                                    'uiAmountString': '1',
                                },
                            },
                            'type': 'account',
                        },
                        'program': 'spl-token',
                        'space': 165,
                    },
                    'executable': False,
                    'lamports': 2039280,
                    'owner': 'TokenkegQfeZyiNwAJbNbGKPFXCWuBvf9Ss623VQ5DA',
                    'rentEpoch': 332,
                },
                'pubkey': 'FxNFQw1sdtuhYtbhnd9HfUabDoW9rrTPuDFkzTQv7DQi',
            },
            coin=Coin(
                symbol='COIN1',
                name='unknown',
                decimals=0,
                blockchain=Blockchain.SOLANA,
                address='addr1',
                standards=None,
                protocol_id=None,
                info=CoinInfo(tags=["tags", "test"]),
            ),
            asset_type=AssetType.AVAILABLE,
            last_updated=None,
            protocol=None,
            is_wallet=True,
        ),
        BalanceItem(
            balance=Decimal('11'),
            balance_raw=Decimal('11'),
            raw={
                'account': {
                    'data': {
                        'parsed': {
                            'info': {
                                'isNative': False,
                                'mint': 'HEL6KGUEvwYgTtcjenf9qeAb2Zg9Yr77usWPY9UZvoQj',
                                'owner': 'FEeSRuEDk8ENZbpzXjn4uHPz3LQijbeKRzhqVr5zPSJ9',
                                'state': 'initialized',
                                'tokenAmount': {
                                    'amount': '11',
                                    'decimals': 0,
                                    'uiAmount': 11.0,
                                    'uiAmountString': '11',
                                },
                            },
                            'type': 'account',
                        },
                        'program': 'spl-token',
                        'space': 165,
                    },
                    'executable': False,
                    'lamports': 2039280,
                    'owner': 'TokenkegQfeZyiNwAJbNbGKPFXCWuBvf9Ss623VQ5DA',
                    'rentEpoch': 333,
                },
                'pubkey': '3XwuM1fhKPmH9onVgtTu8SwC7pPxnKWkUYqqks7NaWPF',
            },
            coin=Coin(
                symbol='COIN2',
                name='unknown',
                decimals=0,
                blockchain=Blockchain.SOLANA,
                address='addr2',
                standards=None,
                protocol_id=None,
                info=None,
            ),
            asset_type=AssetType.AVAILABLE,
            last_updated=None,
            protocol=None,
            is_wallet=True,
        ),
    ]


def test_merge_balances_with_different_coins(solana_api, balances_with_different_coins):
    merged = solana_api.merge_balances_with_same_coin(balances_with_different_coins)
    assert len(merged) == 2
    assert merged == balances_with_different_coins


@pytest.fixture()
def balances_with_mixed_coins():
    return [
        BalanceItem(
            balance=Decimal('1'),
            balance_raw=Decimal('1'),
            raw={},
            coin=Coin(
                symbol='COIN1',
                name='unknown',
                decimals=0,
                blockchain=Blockchain.SOLANA,
                address='addr1',
                standards=None,
                protocol_id=None,
                info=None,
            ),
            asset_type=AssetType.AVAILABLE,
            last_updated=None,
            protocol=None,
            is_wallet=True,
        ),
        BalanceItem(
            balance=Decimal('11'),
            balance_raw=Decimal('11'),
            raw={},
            coin=Coin(
                symbol='COIN2',
                name='unknown',
                decimals=0,
                blockchain=Blockchain.SOLANA,
                address='addr2',
                standards=None,
                protocol_id=None,
                info=None,
            ),
            asset_type=AssetType.AVAILABLE,
            last_updated=None,
            protocol=None,
            is_wallet=True,
        ),
        BalanceItem(
            balance=Decimal('1'),
            balance_raw=Decimal('1'),
            raw={"raw1": "raw1"},
            coin=Coin(
                symbol='unknown',
                name='unknown',
                decimals=0,
                blockchain=Blockchain.SOLANA,
                address='HEL6KGUEvwYgTtcjenf9qeAb2Zg9Yr77usWPY9UZvoQj',
                standards=None,
                protocol_id=None,
                info=None,
            ),
            asset_type=AssetType.AVAILABLE,
            last_updated=None,
            protocol=None,
            is_wallet=True,
        ),
        BalanceItem(
            balance=Decimal('11'),
            balance_raw=Decimal('11'),
            raw={"raw2": "raw2"},
            coin=Coin(
                symbol='unknown',
                name='unknown',
                decimals=0,
                blockchain=Blockchain.SOLANA,
                address='HEL6KGUEvwYgTtcjenf9qeAb2Zg9Yr77usWPY9UZvoQj',
                standards=None,
                protocol_id=None,
                info=None,
            ),
            asset_type=AssetType.AVAILABLE,
            last_updated=None,
            protocol=None,
            is_wallet=True,
        ),
        BalanceItem(
            balance=Decimal('99'),
            balance_raw=Decimal('99'),
            raw={},
            coin=Coin(
                symbol='COIN3',
                name='unknown',
                decimals=0,
                blockchain=Blockchain.SOLANA,
                address='addr3',
                standards=None,
                protocol_id=None,
                info=None,
            ),
            asset_type=AssetType.AVAILABLE,
            last_updated=None,
            protocol=None,
            is_wallet=True,
        ),
        BalanceItem(
            balance=Decimal('11'),
            balance_raw=Decimal('11'),
            raw={"raw3": "raw3"},
            coin=Coin(
                symbol='unknown',
                name='unknown',
                decimals=0,
                blockchain=Blockchain.SOLANA,
                address='HEL6KGUEvwYgTtcjenf9qeAb2Zg9Yr77usWPY9UZvoQj',
                standards=None,
                protocol_id=None,
                info=None,
            ),
            asset_type=AssetType.AVAILABLE,
            last_updated=None,
            protocol=None,
            is_wallet=True,
        ),
    ]


def test_merge_balances_with_different_mixed_coins(
    solana_api, balances_with_mixed_coins
):
    merged = solana_api.merge_balances_with_same_coin(balances_with_mixed_coins)
    assert len(merged) == len(balances_with_mixed_coins) - 2  # 3 balances merged into 1

    for merged_item in merged:
        if not merged_item.raw.get("merged"):
            # skip if not our "merged" item
            continue

        assert merged_item.raw == {
            "merged": [
                balances_with_mixed_coins[2].raw,
                balances_with_mixed_coins[3].raw,
                balances_with_mixed_coins[5].raw,
            ]
        }


@pytest.mark.vcr()
@pytest.mark.integration
def test_get_balance(solana_api):
    balances = solana_api.get_balance('FEeSRuEDk8ENZbpzXjn4uHPz3LQijbeKRzhqVr5zPSJ9')
    assert len(balances) > 200

    merged_balance = []
    for item in balances:
        if item.coin.address == 'HEL6KGUEvwYgTtcjenf9qeAb2Zg9Yr77usWPY9UZvoQj':
            merged_balance.append(item)

    assert len(merged_balance) == 1
    merged_balance = merged_balance[0]
    assert len(merged_balance.raw.get("merged")) == 2


def test_use_custom_url():
    api = SolanaApi('https://proxy/solana/')
    assert api.api_options.base_url == 'https://proxy/solana/'


def test_use_base_url():
    api = SolanaApi()
    assert api.API_BASE_URL
    assert api.api_options.base_url == api.API_BASE_URL



from decimal import Decimal

from pytest import mark

from blockapi.api.terra_money import TerraMoneyApi
from blockapi.test_data import test_addresses


class TestTerraMoneyApi:
    ADDRESS = test_addresses["LUNA"][0]

    def test_init(self):
        api = TerraMoneyApi(address=self.ADDRESS)
        assert api

    @mark.vcr()
    def test_get_balance(self):
        api = TerraMoneyApi(address=self.ADDRESS)
        result = api.get_balance()

        assert result == [
            {'symbol': 'LUNA', 'amount': Decimal('34682.501828')},
            {'symbol': 'KRT', 'amount': Decimal('9412.963717')},
            {'symbol': 'SDT', 'amount': Decimal('0.052803')},
            {'symbol': 'UST', 'amount': Decimal('0.012112')},
            {'symbol': 'MNT', 'amount': Decimal('1122.501887')},
        ]

    @mark.vcr()
    def test_get_txs(self):
        api = TerraMoneyApi(address=self.ADDRESS)
        result = api.get_txs()

        assert len(result) == 11

    @mark.vcr()
    def test_get_delegations(self):
        api = TerraMoneyApi(address=self.ADDRESS)
        result = api.get_delegations()

        assert result



from decimal import Decimal

import pytest

from blockapi.test.v2.api.conftest import read_file
from blockapi.v2.api.subscan import PolkadotSubscanApi


@pytest.mark.parametrize(
    'address, response_path, expected_balance',
    [
        (
            '1NjjK81oA9is7eWhFJ7n7kQhaPT3pnxLFH5MkwDfduAiWE9',
            'data/subscan_polkadot_response_WE9.json',
            Decimal('4649.9139999889'),
        ),
        (
            '15j4dg5GzsL1bw2U2AWgeyAk6QTxq43V7ZPbXdAmbVLjvDCK',
            'data/subscan_polkadot_response_DCK.json',
            Decimal('49296244.7887007465'),
        ),
    ],
)
def test_fetch_balances(requests_mock, address, response_path, expected_balance):
    requests_mock.post(
        'https://polkadot.api.subscan.io/api/v2/scan/search',
        text=read_file(response_path),
    )

    api = PolkadotSubscanApi()
    balances = api.get_balance(address)
    assert sum(b.balance for b in balances) == expected_balance



from typing import Dict, Iterable, List, Optional

from eth_utils import to_checksum_address

from blockapi.v2.base import ApiOptions, BlockchainApi, IBalance
from blockapi.v2.coins import COIN_ETH
from blockapi.v2.models import BalanceItem, Blockchain, Coin, CoinInfo


class EthplorerApi(BlockchainApi, IBalance):
    """
    Ethereum
    API docs: https://github.com/EverexIO/Ethplorer/wiki/Ethplorer-API
    Explorer: https://ethplorer.io
    """

    coin = COIN_ETH
    api_options = ApiOptions(
        blockchain=Blockchain.ETHEREUM,
        base_url='https://api.ethplorer.io',
        rate_limit=0.2,  # 0.1 in case of api_key
    )

    supported_requests = {'get_info': '/getAddressInfo/{address}?apiKey={api_key}'}

    def __init__(self, api_key: str = 'freekey'):
        super().__init__(api_key)

    def get_balance(self, address: str) -> List[BalanceItem]:
        response = self.get('get_info', address=address, api_key=self.api_key)

        balances = []

        _eth = self._parse_eth_balance(response)
        if _eth is not None:
            balances.append(_eth)

        balances.extend(list(self._parse_token_balances(response)))

        return balances

    def _parse_eth_balance(self, response: Dict) -> Optional[BalanceItem]:
        _eth_raw = response['ETH']
        if int(_eth_raw['rawBalance']) == 0:
            return

        return BalanceItem.from_api(
            balance_raw=_eth_raw['rawBalance'],
            coin=self.coin,
            last_updated=None,
            raw=_eth_raw,
        )

    def _parse_token_balances(self, response: Dict) -> Iterable[BalanceItem]:
        for _token_raw in response.get('tokens', []):
            if _token_raw.get('rawBalance') is None or _token_raw['rawBalance'] == 0:
                continue

            info = _token_raw['tokenInfo']
            coin = Coin.from_api(
                blockchain=self.api_options.blockchain,
                decimals=info.get('decimals', 0),
                symbol=info.get('symbol'),
                name=info.get('name'),
                address=to_checksum_address(info['address']),
                standards=None,  # parse from tags?
                info=CoinInfo.from_api(
                    tags=info.get('publicTags'),
                    total_supply=info.get('totalSupply'),
                    logo_url=(
                        self._format_logo_url(info.get('image'))
                        if info.get('image')
                        else None
                    ),
                    coingecko_id=info.get('coingecko'),
                    website=info.get('website'),
                ),
            )

            yield BalanceItem.from_api(
                balance_raw=_token_raw['rawBalance'],
                coin=coin,
                last_updated=info.get('lastUpdated'),
                raw=_token_raw,
            )

    @staticmethod
    def _format_logo_url(raw_url: str) -> str:
        return f'https://ethplorer.io{raw_url}'



from decimal import Decimal

import pytest

from blockapi.test.v2.api.conftest import read_file
from blockapi.v2.api.blockchair import BlockchairDogecoinApi

# noinspection SpellCheckingInspection
doge_test_address = 'DH5yaieqoZN36fDVciNyRueRGvGLR3mr7L'


def test_fetch_balances(requests_mock, doge_balance_response):
    requests_mock.get(
        f'https://api.blockchair.com/dogecoin/dashboards/address/{doge_test_address}?limit=10,0&offset=0,0',
        text=doge_balance_response,
    )

    api = BlockchairDogecoinApi()
    balances = api.get_balance(doge_test_address)
    assert len(balances) == 1
    assert balances[0].balance == Decimal('50285.15921524')


def test_fetch_transactions(
    requests_mock, doge_balance_response, doge_transactions_response
):
    requests_mock.get(
        f'https://api.blockchair.com/dogecoin/dashboards/address/{doge_test_address}?limit=10,0&offset=0,0',
        text=doge_balance_response,
    )

    # noinspection SpellCheckingInspection
    addresses = [
        "ccb2cd37527f725f06ae19f4e46881dc2ac4c5072d9de40a2c563f61536aa7ed",
    ]

    requests_mock.get(
        'https://api.blockchair.com/dogecoin/dashboards/transactions/'
        + ','.join(addresses),
        text=doge_transactions_response,
    )

    api = BlockchairDogecoinApi()

    txs = api.get_transactions(doge_test_address)
    assert len(txs) == 1
    assert txs[0].fee == Decimal('0')
    assert len(txs[0].operations) == 1
    assert txs[0].operations[0].amount == Decimal('31.28571429')


@pytest.fixture()
def doge_transactions_response():
    return read_file('data/blockchair_doge_transaction_response.json')


@pytest.fixture
def doge_balance_response():
    return read_file('data/blockchair_doge_balance_response.json')



from decimal import Decimal

import pytest

from blockapi.test.v2.api.conftest import read_file
from blockapi.v2.api.blockchainos import BlockchainosApi

# noinspection SpellCheckingInspection
test_address = 'GCPQQIX2LRX2J63C7AHWDXEMNGMZR2UI2PRN5TCSOVMEMF7BAUADMKH5'


@pytest.fixture
def api():
    return BlockchainosApi()


def test_fetch_balances(requests_mock, api, bos_balance_response):
    requests_mock.get(
        f'https://mainnet.blockchainos.org/api/v1/accounts/{test_address}',
        text=bos_balance_response,
    )

    balances = api.get_balance(test_address)
    assert len(balances) == 1
    assert balances[0].balance == Decimal('1012356993.757')
    assert balances[0].balance == balances[0].balance_raw * Decimal('1e-7')


def test_fetch_transactions(
    requests_mock, api, bos_transactions_response, bos_operations_responses
):
    create_transaction_mocks(
        bos_operations_responses, bos_transactions_response, requests_mock
    )

    txs = api.get_transactions(test_address, limit=3)
    assert len(txs) == 3
    assert txs[0].fee == Decimal('0')
    assert len(txs[0].operations) == 2
    assert txs[0].operations[1].amount == Decimal('50')


def create_transaction_mocks(
    bos_operations_responses, bos_transactions_response, requests_mock
):
    requests_mock.get(
        f'https://mainnet.blockchainos.org/api/v1/accounts/{test_address}/transactions?limit=3&reverse=true',
        text=bos_transactions_response,
    )
    for hash, responses in bos_operations_responses.items():
        for index in range(len(responses)):
            requests_mock.get(
                f'https://mainnet.blockchainos.org/api/v1/transactions/{hash}/operations/{index}',
                text=responses[index],
            )


@pytest.fixture()
def bos_transactions_response():
    return read_file('data/bos_transaction_response.json')


@pytest.fixture
def bos_balance_response():
    return read_file('data/bos_balance_response.json')


@pytest.fixture()
def bos_operations_responses():
    # noinspection SpellCheckingInspection
    return {
        '8tok9mHjGURhcnbLkrhcEDgQoXvoYjLVs2ipxAAsi2d6': [
            read_file('data/bos_operations_response_8t_1.json'),
            read_file('data/bos_operations_response_8t_2.json'),
        ],
        "R1nddzPKA5uh5a7ZTjS113aUth9t1Bv68hjjjxGCqgS": [
            read_file('data/bos_operations_response_r1_1.json'),
            read_file('data/bos_operations_response_r1_2.json'),
        ],
    }



rewards_abi = [
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "address",
                "name": "_claimant",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "_balance",
                "type": "uint256",
            },
        ],
        "name": "Claimed",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "previousOwner",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "newOwner",
                "type": "address",
            },
        ],
        "name": "OwnershipTransferred",
        "type": "event",
    },
    {
        "inputs": [],
        "name": "candidate",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "_liquidityProvider",
                "type": "address",
            },
            {"internalType": "uint256", "name": "_begin", "type": "uint256"},
            {"internalType": "uint256", "name": "_end", "type": "uint256"},
        ],
        "name": "claimStatus",
        "outputs": [{"internalType": "bool[]", "name": "", "type": "bool[]"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "_account", "type": "address"},
            {"internalType": "uint256", "name": "_week", "type": "uint256"},
            {
                "internalType": "uint256",
                "name": "_claimedBalance",
                "type": "uint256",
            },
            {
                "internalType": "bytes32[]",
                "name": "_merkleProof",
                "type": "bytes32[]",
            },
        ],
        "name": "claimWeek",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "_account", "type": "address"},
            {
                "components": [
                    {
                        "internalType": "uint256",
                        "name": "week",
                        "type": "uint256",
                    },
                    {
                        "internalType": "uint256",
                        "name": "balance",
                        "type": "uint256",
                    },
                    {
                        "internalType": "bytes32[]",
                        "name": "merkleProof",
                        "type": "bytes32[]",
                    },
                ],
                "internalType": "struct MerkleRedeemUpgradeSafe.Claim[]",
                "name": "_claims",
                "type": "tuple[]",
            },
        ],
        "name": "claimWeeks",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "uint256", "name": "", "type": "uint256"},
            {"internalType": "address", "name": "", "type": "address"},
        ],
        "name": "claimed",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "getLengthOfMerkleRoots",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {
                "internalType": "contract IERC20",
                "name": "_token",
                "type": "address",
            },
            {
                "internalType": "uint256",
                "name": "_vestingPeriod",
                "type": "uint256",
            },
        ],
        "name": "initialize",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "name": "merkleRootIndexes",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "name": "merkleRootTimestampMap",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "uint256", "name": "_begin", "type": "uint256"},
            {"internalType": "uint256", "name": "_end", "type": "uint256"},
        ],
        "name": "merkleRoots",
        "outputs": [{"internalType": "bytes32[]", "name": "", "type": "bytes32[]"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "owner",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "renounceOwnership",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "uint256", "name": "_week", "type": "uint256"},
            {
                "internalType": "bytes32",
                "name": "_merkleRoot",
                "type": "bytes32",
            },
            {
                "internalType": "uint256",
                "name": "_totalAllocation",
                "type": "uint256",
            },
        ],
        "name": "seedAllocations",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "address", "name": "newOwner", "type": "address"}],
        "name": "setOwner",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "token",
        "outputs": [{"internalType": "contract IERC20", "name": "", "type": "address"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "updateOwner",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "_liquidityProvider",
                "type": "address",
            },
            {"internalType": "uint256", "name": "_week", "type": "uint256"},
            {
                "internalType": "uint256",
                "name": "_claimedBalance",
                "type": "uint256",
            },
            {
                "internalType": "bytes32[]",
                "name": "_merkleProof",
                "type": "bytes32[]",
            },
        ],
        "name": "verifyClaim",
        "outputs": [{"internalType": "bool", "name": "valid", "type": "bool"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "vestingPeriod",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "name": "weekMerkleRoots",
        "outputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
        "stateMutability": "view",
        "type": "function",
    },
]



from decimal import Decimal

import pytest

from blockapi.test.v2.api.conftest import read_file
from blockapi.v2.api.blockchair import BlockchairBitcoinApi

# noinspection SpellCheckingInspection
btc_test_address = '35hK24tcLEWcgNA4JxpvbkNkoAcDGqQPsP'


def test_fetch_balances(requests_mock, btc_balance_response):
    requests_mock.get(
        f'https://api.blockchair.com/bitcoin/dashboards/address/{btc_test_address}?limit=10,0&offset=0,0',
        text=btc_balance_response,
    )

    api = BlockchairBitcoinApi()
    balances = api.get_balance(btc_test_address)
    assert len(balances) == 1
    assert balances[0].balance == Decimal('0.00064363')


def test_fetch_transactions(
    requests_mock, btc_balance_response, btc_transactions_response
):
    requests_mock.get(
        f'https://api.blockchair.com/bitcoin/dashboards/address/{btc_test_address}?limit=3,0&offset=0,0',
        text=btc_balance_response,
    )

    # noinspection SpellCheckingInspection
    transactions = [
        "3fcb8ac40fd7e1bfe2a95b6704d8af3bbd88640c0177ec118e26e06d3a06cc07",
        "e39457afee2730fa3eac9016efa02dac1c28998b60e87efdaa604ed0d175567f",
        "21cc75fd255111ebc9a4c81ce08c45f26e8d8721b4a8912452648ca9ff9d54c1",
    ]

    requests_mock.get(
        'https://api.blockchair.com/bitcoin/dashboards/transactions/'
        + ','.join(transactions),
        text=btc_transactions_response,
    )

    api = BlockchairBitcoinApi()

    txs = api.get_transactions(btc_test_address, limit=3)
    assert len(txs) == 1
    assert txs[0].fee == Decimal('0.00056963')
    assert len(txs[0].operations) == 1
    assert txs[0].operations[0].amount == Decimal('0.20087268')


@pytest.fixture()
def btc_transactions_response():
    return read_file('data/blockchair_btc_transaction_response.json')


@pytest.fixture
def btc_balance_response():
    return read_file('data/blockchair_btc_balance_response.json')



import pytest

from blockapi.test.v2.test_data import (
    BAD_ADDRESSES,
    DEBANK_API_KEY,
    NON_EMPTY_VALID_ADDRESSES_BY_SYMBOL,
    get_debank_addresses,
    yield_all_api_classes,
    yield_api_instances,
)
from blockapi.v2.api.debank import DebankApi
from blockapi.v2.api.optimistic_etherscan import OptimismEtherscanApi
from blockapi.v2.base import ApiException
from blockapi.v2.models import AssetType, BalanceItem, Blockchain


@pytest.mark.integration
@pytest.mark.parametrize('api_instance', yield_api_instances())
def test_get_balance_for_valid_non_empty_address(api_instance):
    for address in NON_EMPTY_VALID_ADDRESSES_BY_SYMBOL.get(
        api_instance.coin.symbol, []
    ):
        balances = api_instance.get_balance(address)

        assert all(isinstance(b, BalanceItem) for b in balances)
        assert len(balances) > 0


@pytest.mark.integration
@pytest.mark.parametrize('api_instance', yield_api_instances())
def test_get_balance_for_bad_address(api_instance):
    if isinstance(api_instance, OptimismEtherscanApi):
        # OptimismEtherscanApi does not validate addresses, so we skip it.
        return

    for address in BAD_ADDRESSES:
        with pytest.raises(ApiException):
            api_instance.get_balance(address)


@pytest.mark.integration
@pytest.mark.parametrize('address', get_debank_addresses())
def test_get_balance_for_debank(address):
    api_instance = DebankApi(DEBANK_API_KEY, False)
    balances = api_instance.get_balance(address)

    assert all(isinstance(b, BalanceItem) for b in balances)
    assert all(AssetType(b.asset_type) is not None for b in balances)
    assert all(Blockchain(b.coin.blockchain) is not None for b in balances)
    assert len(balances) > 0


@pytest.mark.integration
@pytest.mark.parametrize('address', get_debank_addresses())
def test_get_portfolio_for_debank(address):
    api_instance = DebankApi(DEBANK_API_KEY, False)
    pools = api_instance.get_portfolio(address)

    balances = [item for pool in pools for item in pool.items]

    assert all(isinstance(b, BalanceItem) for b in balances)
    assert all(AssetType(b.asset_type) is not None for b in balances)
    assert all(Blockchain(b.coin.blockchain) is not None for b in balances)
    assert len(balances) > 0


@pytest.mark.parametrize('api_instance', yield_all_api_classes())
def test_verify_repr_for_instance(api_instance):
    assert repr(api_instance)



import os

import pytest

from blockapi.v2.api.perpetual import PerpetualApi, perp_contract_address


@pytest.fixture
def perp_api():
    return PerpetualApi('http://localhost:2048/')


test_address = '0x134089B387E22f52b1e06CC80d9a5F622032EF74'


def test_perp_contract_address():
    contract = perp_contract_address('PERP')
    assert contract == '0xbC396689893D065F41bc2C6EcbeE5e0085233447'


def test_perp_invalid_contract_raises():
    with pytest.raises(ValueError, match='Invalid contract name.'):
        perp_contract_address("abc")


def test_perp_has_coin(perp_api):
    assert perp_api.coin.symbol == 'PERP'


def filter_infura_key(request):
    if 'infura.io' in request.host:
        request.uri = 'https://mainnet.infura.io/v3/API_KEY_FILTERED'
    return request


@pytest.mark.integration
@pytest.mark.vcr(before_record_request=filter_infura_key)
def test_perp_get_balances():
    key = os.environ.get('INFURA_API_KEY')
    perp_api = PerpetualApi(f'https://mainnet.infura.io/v3/{key}')
    balances = perp_api.get_balance(test_address)
    assert balances[0].balance
    assert balances[1].balance



import inspect
import os

import pytest

from blockapi.v2.api.blockchair import BlockchairBitcoinApi
from blockapi.v2.api.cosmos import CosmosApi, CosmosApiBase
from blockapi.v2.api.covalenth.arbitrum import ArbitrumCovalentApi
from blockapi.v2.api.covalenth.astar import AstarCovalentApi
from blockapi.v2.api.covalenth.avalanche import AvalancheCovalentApi
from blockapi.v2.api.covalenth.axie import AxieCovalentApi
from blockapi.v2.api.covalenth.base import CovalentApiBase
from blockapi.v2.api.covalenth.binance_smart_chain import BscCovalentApi
from blockapi.v2.api.covalenth.fantom import FantomCovalentApi
from blockapi.v2.api.covalenth.heco import HECOCovalentApi
from blockapi.v2.api.covalenth.iotex import IoTEXCovalentApi
from blockapi.v2.api.covalenth.klaytn import KlaytnCovalentApi
from blockapi.v2.api.covalenth.moonbeam import MoonBeamCovalentApi
from blockapi.v2.api.covalenth.palm import PalmCovalentApi
from blockapi.v2.api.covalenth.polygon import PolygonCovalentApi
from blockapi.v2.api.covalenth.rsk import RskCovalentApi
from blockapi.v2.api.debank import DebankApi
from blockapi.v2.api.ethplorer import EthplorerApi
from blockapi.v2.api.optimistic_etherscan import OptimismEtherscanApi
from blockapi.v2.api.perpetual import PerpetualApi
from blockapi.v2.api.solana import SolanaApi
from blockapi.v2.api.subscan import KusamaSubscanApi, PolkadotSubscanApi, SubscanApi
from blockapi.v2.api.synthetix import SynthetixApi
from blockapi.v2.api.terra import TerraApi

# TODO create method for auto loading all classes
from blockapi.v2.base import IBalance
from blockapi.v2.coins import (
    COIN_ATOM,
    COIN_AVAX,
    COIN_BNB,
    COIN_DOT,
    COIN_ETH,
    COIN_FTM,
    COIN_HT,
    COIN_IOTX,
    COIN_KLAY,
    COIN_KSM,
    COIN_MATIC,
    COIN_MOVR,
    COIN_PALM,
    COIN_RON,
    COIN_RSK,
    COIN_SDN,
    COIN_SOL,
    COIN_TERRA,
)

COVALENT_API_KEY = os.getenv("COVALENT_API_KEY")
DEBANK_API_KEY = os.getenv("DEBANK_API_KEY")

API_CLASSES = [
    EthplorerApi,
    SolanaApi,
    TerraApi,
    ArbitrumCovalentApi,
    AstarCovalentApi,
    AvalancheCovalentApi,
    AxieCovalentApi,
    BscCovalentApi,
    EthplorerApi,
    FantomCovalentApi,
    HECOCovalentApi,
    IoTEXCovalentApi,
    KlaytnCovalentApi,
    MoonBeamCovalentApi,
    PalmCovalentApi,
    PolygonCovalentApi,
    RskCovalentApi,
    OptimismEtherscanApi,
    DebankApi,
    PolkadotSubscanApi,
    KusamaSubscanApi,
    CosmosApi,
    SynthetixApi,
    PerpetualApi,
    BlockchairBitcoinApi,
]

NON_EMPTY_VALID_ADDRESSES_BY_SYMBOL = {
    COIN_ETH.symbol: [
        '0xeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeee',
    ],
    COIN_SOL.symbol: [
        '31dpiondDhZaqK23Re8kzkhY6CFEG9ZTQnr3shQm7g8b',
    ],
    COIN_TERRA.symbol: [
        'terra1yltenl48mhl370ldpyt83werd9x3s645509gaf',
    ],
    COIN_SDN.symbol: [
        '0xfc43f5f9dd45258b3aff31bdbe6561d97e8b71de',
    ],
    COIN_AVAX.symbol: [
        '0xfc43f5f9dd45258b3aff31bdbe6561d97e8b71de',
    ],
    COIN_RON.symbol: [
        '0xfc43f5f9dd45258b3aff31bdbe6561d97e8b71de',
    ],
    COIN_BNB.symbol: [
        '0xfc43f5f9dd45258b3aff31bdbe6561d97e8b71de',
    ],
    COIN_FTM.symbol: [
        '0xfc43f5f9dd45258b3aff31bdbe6561d97e8b71de',
    ],
    COIN_HT.symbol: [
        '0xfc43f5f9dd45258b3aff31bdbe6561d97e8b71de',
    ],
    COIN_IOTX.symbol: [
        '0xfc43f5f9dd45258b3aff31bdbe6561d97e8b71de',
    ],
    COIN_KLAY.symbol: [
        '0xfc43f5f9dd45258b3aff31bdbe6561d97e8b71de',
    ],
    COIN_MOVR.symbol: [
        '0xfc43f5f9dd45258b3aff31bdbe6561d97e8b71de',
    ],
    COIN_PALM.symbol: [
        '0xfc43f5f9dd45258b3aff31bdbe6561d97e8b71de',
    ],
    COIN_MATIC.symbol: [
        '0xfc43f5f9dd45258b3aff31bdbe6561d97e8b71de',
    ],
    COIN_RSK.symbol: [
        '0xfc43f5f9dd45258b3aff31bdbe6561d97e8b71de',
    ],
    COIN_DOT.symbol: ['15j4dg5GzsL1bw2U2AWgeyAk6QTxq43V7ZPbXdAmbVLjvDCK'],
    COIN_ATOM.symbol: ['cosmos1r4v9t46zyu6df0jwtmtpn0pq864dpn7c5sha5u'],
    COIN_KSM.symbol: ['EK8HP4biJ8FBktaREBd9Bt3E85QsVTmpYzn5aou7iiUDDgB'],
}

BAD_ADDRESSES = [
    '1',
    '10000',
    'AAAAA',
    '0x123',
    'ýžýščš&^^~',
    '0xAF3E9eF83B1dc9Db18cD5923e3112F6Cd8bfDAeX',
    '0xAF3E9eF83B1dc9Db18cD5923e3112F6Cd8bfDAeDD',
]


def yield_api_instances():
    for api_cls in yield_covalent_api_classes():
        yield _pytest_param(api_cls(api_key=COVALENT_API_KEY))

    for api_cls in yield_api_ibalance_classes():
        yield _pytest_param(api_cls())


def _pytest_param(api_inst):
    return pytest.param(api_inst, id=str(api_inst))


def yield_covalent_api_classes():
    return CovalentApiBase.__subclasses__()


def yield_api_ibalance_classes():
    return [
        x
        for x in IBalance.__subclasses__()
        if not issubclass(
            x,
            (
                CosmosApiBase,
                CovalentApiBase,
                DebankApi,
                PerpetualApi,
                SynthetixApi,
                SubscanApi,
            ),
        )
        and not inspect.isabstract(x)
    ]


def yield_all_api_classes():
    return [x for x in IBalance.__subclasses__()]


def get_debank_addresses():
    result = []
    for key, items in NON_EMPTY_VALID_ADDRESSES_BY_SYMBOL.items():
        if key in (
            COIN_TERRA.symbol,
            COIN_SOL.symbol,
            COIN_DOT.symbol,
            COIN_KSM.symbol,
            COIN_ATOM.symbol,
        ):
            continue

        for item in items:
            result.append(pytest.param(item, id=key))

    return result



from blockapi.services import BlockchainAPI


class EthplorerAPI(BlockchainAPI):
    """
    Ethereum
    API docs: https://github.com/EverexIO/Ethplorer/wiki/Ethplorer-API
    Explorer: https://ethplorer.io
    """

    symbol = 'ETH'
    base_url = 'http://api.ethplorer.io'
    default_api_key = 'freekey'
    rate_limit = 0.2
    coef = 1e-18
    start_offset = None
    max_items_per_page = None
    page_offset_step = None

    supported_requests = {'get_info': '/getAddressInfo/{address}?apiKey={api_key}'}

    def __init__(self, address, api_key=None):
        if not api_key:
            api_key = self.default_api_key

        super().__init__(address, api_key)

        # cache info
        self._info = None

        if self.api_key != self.default_api_key:
            self.rate_limit = 0.1

    def get_balance(self):
        balances = []
        response = self.info

        if response.get('ETH', {}).get('balance'):
            balances.append(
                {'symbol': self.symbol, 'amount': response['ETH']['balance']}
            )

        for token in response.get('tokens', []):
            info = token['tokenInfo']
            decimals = int(info['decimals']) if info.get('decimals') else 18

            balances.append(
                {
                    'symbol': info.get('symbol', 'unknown'),
                    'address': info['address'],
                    'amount': token['balance'] * pow(10, -decimals),
                    'name': info.get('name', 'Unknown'),
                }
            )

        if not balances:
            return None

        return balances

    @property
    def info(self):
        if not self._info:
            self._info = self.request(
                'get_info', address=self.address, api_key=self.api_key
            )
        return self._info
